//
//  BlockBlockKext.c
//  BlockBlockKext
//
//  Created by Patrick Wardle on 11/18/15.
//  Copyright © 2015 Objective-See. All rights reserved.
//

#include <sys/proc.h>
#include <sys/kauth.h>
#include <sys/kern_event.h>
#include <libkern/libkern.h>
#include <mach/mach_types.h>


/* FUNCTIONS */

//start
// ->register kauth listener
kern_return_t start(kmod_info_t * ki, void *d);

//stop
// ->unregsiter kauth listener
kern_return_t stop(kmod_info_t *ki, void *d);

//kauth callback
// ->for KAUTH_FILEOP_EXEC events, broadcast process notifications to user-mode
static int processExec(kauth_cred_t credential, void* idata, kauth_action_t action, uintptr_t arg0, uintptr_t arg1, uintptr_t arg2, uintptr_t arg3);

/* GLOBALS */

//kauth listener
// ->scope KAUTH_SCOPE_FILEOP
kauth_listener_t kauthListener = NULL;

/* CODE */

//kauth callback
// ->for KAUTH_FILEOP_EXEC events, broadcast process notifications to user-mode
static int processExec(kauth_cred_t credential, void* idata, kauth_action_t action, uintptr_t arg0, uintptr_t arg1, uintptr_t arg2, uintptr_t arg3)
{
    //path
    char path[MAXPATHLEN] = {0};
    
    //uid
    uid_t uid = -1;
    
    //pid
    pid_t pid = -1;
    
    //ppid
    pid_t ppid = -1;
    
    //ignore all non exec events
    if(KAUTH_FILEOP_EXEC != action)
    {
        //bail
        goto bail;
    }
    
    //path is arg1 (per sys/k_auth.h)
    // ->make copy, so boardcast to usermode works
    strncpy(path, (const char*)arg1, MAXPATHLEN-1);
    
    //get UID
    uid = kauth_getuid();
    
    //get pid
    pid = proc_selfpid();
    
    //get ppid
    ppid = proc_selfppid();
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: new process: %s %d/%d/%d\n", path, pid, ppid, uid);

    //TODO: send to user mode!
    
    
//bail
bail:
    
    return KAUTH_RESULT_DEFER;
}

//start function, automatically invoked
// ->register KAuth listener
kern_return_t start(kmod_info_t * ki, void *d)
{
    //status var
    kern_return_t status = KERN_FAILURE;
    
    //dbg msg
    printf("BLOCKBLOCK KEXT: starting...\n");
    
    //register listener
    // ->scope 'KAUTH_SCOPE_FILEOP'
    kauthListener = kauth_listen_scope(KAUTH_SCOPE_FILEOP, &processExec, NULL);
    if(NULL == kauthListener)
    {
        //err msg
        printf("BLOCKBLOCK KEXT ERROR: kauth_listen_scope('KAUTH_SCOPE_VNODE',...) failed\n");
        
        //bail
        goto bail;
    }
    
    //happy
    status = KERN_SUCCESS;
    
//bail
bail:
    
    return status;
}

//stop function, automatically invoked
// ->should never get called in release mode
kern_return_t stop(kmod_info_t *ki, void *d)
{
    //status
    kern_return_t status = KERN_FAILURE;
    
    //unregister listener
    if(NULL != kauthListener)
    {
        //unregister
        kauth_unlisten_scope(kauthListener);
        
        //unset
        kauthListener = NULL;
    }
    
    //happy
    status = KERN_SUCCESS;
    
//bail
bail:
    
    return status;
    
}
